---------------------------------------------------------------------------
--                                                                       --
--         Cfg2Dot generates dot output for a CFG using Ast2Cfg          --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--        Copyright (c) 2007-2009 Raul Fechete, Georg Kienesberger       --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of Cfg2Dot.                                         --
--                                                                       --
-- Cfg2Dot is free software; you can redistribute it and/or modify       --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- Cfg2Dot is distributed in the hope that it will be useful,            --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

with Ada.Text_IO; use Ada.Text_IO;
with Ada.Strings.Fixed;
with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with Ada.Directories;
with Ada.Characters.Latin_1;

with Ast2Cfg.Flow_World;
with Ast2Cfg.Id;
with Ast2Cfg.Flow_Types; use Ast2Cfg.Flow_Types;
use Ast2Cfg.Flow_Types.Declaration_Ptr_List;
use Ast2Cfg.Flow_Types.Formal_Decl_Class_Ptr_List;
with Ast2Cfg.Pkgs; use Ast2Cfg.Pkgs;
with Ast2Cfg.CFGs; use Ast2Cfg.CFGs;
with Ast2Cfg.Nodes; use Ast2Cfg.Nodes;
with Ast2Cfg.Utils; use Ast2Cfg.Utils;
use Ast2Cfg;
with Ast2Cfg.Generic_List;

with ast2cfg.output; use ast2cfg.output;

with Cfg_To_Dot.Options;

package body Cfg_To_Dot.Output is

   -- G E N E R A L

   type Cfg_Type is (Normal, Parameters);

   --
   -- Compose the (*.*...*) Name of a Structure Unit (Pkg / CFG)
   --
   function Compose_Hierarchical_Name(Flow: in Flow_Object'Class) return String is
      Prev_Flows: Flow_Class_Ptr_List.Object := Get_Preds(Flow);
      Prev_Flow:  Flow_Class_Ptr             := null;
   begin
      if (Flow_Class_Ptr_List.Count(Prev_Flows) > 1) then
         -- We have more than one Predecessor for this Node!
         Put_Line(Standard_Error, "[DOT/OUTPUT] [ERR] More than one PREV for Flow Object: ");
         -- Display all the Prevs
         Put(Standard_Error, Get_Name(Flow) & ": ");
         Flow_Class_Ptr_List.Reset(Prev_Flows);
         while Flow_Class_Ptr_List.Has_Next(Prev_Flows) loop
            Flow_Class_Ptr_List.Get_Next(Prev_Flows, Prev_Flow);
            Put(Standard_Error, Get_Name(Prev_Flow.all) & ",");
         end loop;
         Put_Line(Standard_Error,"");
         return "";
      elsif (Flow_Class_Ptr_List.Count(Prev_Flows) = 1) then
         -- We have only one Predecessor for this Node, so use its Name
         Flow_Class_Ptr_List.Reset(Prev_Flows);
         Flow_Class_Ptr_List.Get_Next(Prev_Flows, Prev_Flow);
         return Compose_Hierarchical_Name(Prev_Flow.all) & "." & Get_Name(Flow);
      else
         -- No more Predecessors - we have finished!
         return Get_Name(Flow);
      end if;
   end Compose_Hierarchical_Name;

   -- Make a valid Filename out of some String
   -- Returns "" in case it can't be done (because stdout was chosen)
   function Make_Filename (Name: in String) return String is
      Filename: Unbounded_String := To_Unbounded_String(Name);
   begin
      if To_String(Options.Dir) /= "" then
         -- Replace ' 's with '_'s
         for i in 1..Length(Filename) loop
            if Element(Filename, i) = ' ' then
               Replace_Element(Filename, i, '_');
            end if;
         end loop;
         -- Create basic Filename
         Filename := Options.Dir & To_Unbounded_String("/") & Name;
         -- Make sure the Filename is unique and append ".dot"
         declare
            i: Positive := 1;
         begin
            if Ada.Directories.Exists(To_String(Filename) & ".dot") then
               while Ada.Directories.Exists(To_String(Filename) & "." & Ada.Strings.Fixed.Trim(Positive'Image(i), Ada.Strings.Both) & ".dot") loop
                  i := i + 1;
               end loop;
               Append(Filename, "." & Ada.Strings.Fixed.Trim(Positive'Image(i), Ada.Strings.Both) & ".dot");
            else
               Append(Filename, ".dot");
            end if;
         end;
      else
         Filename := To_Unbounded_String("");
      end if;
      return To_String(Filename);
   end Make_Filename;

   -- W O R L D

   -- Forward Declarations

   -- ** General Output Methods

   -- Dump a Hierarchy Subtree
   procedure Tree_Dot_Dump (Flow: in Flow_Object'Class);
   -- Dump a Hierarchy Subtree recursively
   procedure Tree_Dot_Dump_Rec (Flow: in Flow_Object'Class);
   -- Dump the complete Content of a Package Flow Structure recursively
   procedure Pkg_Dot_Dump_Rec (Flow: in Flow_Object'Class);

   -- ** Special List Output Methods

   -- Generate a String out of a References List (with, use, ..)
   function Refs(Name: in String; Refs_List: in Ref_Class_Ptr_List.Object) return String;
   -- Generate a String out of an Alias List (Renamings, Instantiations, ..)
   function Aliases (Name: in String; Alias_List: in Alias_Class_Ptr_List.Object) return String;
   -- Generate a String out of a Declarations List (Variables, Parameters, ..)
   function RDecls(Name: in String; Decl_List: in Declaration_Ptr_List.Object) return String;
   -- Generate a String out of a Formal Declarations List (in Generic Entities)
   function FDecls(Name: in String; Decl_List: in Formal_Decl_Class_Ptr_List.Object) return String;
   -- Generate a String out of a Type Description List (Type Declarations, ..)
   function Types(Name: in String; Types_List: in Type_Descr_Ptr_List.Object) return String;

   -- ** Special Tree Output Methods

   -- Dump a Parameter Tree
   procedure Write_Param_Tree (Prev: in String; New_Name: in String; Root: in Node_Class_Ptr);
   -- Dump the Parameter Trees for a List of Declarations
   procedure Write_Param_Trees (Name: in String; Decl_List: in Declaration_Ptr_List.Object);
   -- Dump the Parameter Trees for a List of Formal Declarations
   procedure Write_Formal_Param_Trees (Name: in String; Decl_List: in Formal_Decl_Class_Ptr_List.Object);
   -- Dump the Parameter Trees for a List of Aliases
   procedure Write_Alias_Param_Trees (Name: in String; Alias_List: in Alias_Class_Ptr_List.Object);
   -- Dump the Parameter Trees for a List of Type Descriptions
   procedure Write_Type_Param_Trees (Name: in String; Types_List: in Type_Descr_Ptr_List.Object);

   -- ** Special Structure Methods

   -- Construct a String Image of a Type
   function Type_Image (Type_Info: in Type_Descr) return String;

   -- ** Flow Structure Methods

   procedure Write_Node (Cur: in out Node_Object'Class; Cfg_Class: in Cfg_Type);

   --
   -- Output everything in the Flow World
   --
   procedure World_Dot_Dump (World: in Flow_World.World_Object) is
      Pkg:  Pkg_Class_Ptr             := null;
      Pkgs: Pkg_Class_Ptr_List.Object := Flow_World.Get_Pkgs(World);
   begin
      Pkg_Class_Ptr_List.Reset(Pkgs);
      while Pkg_Class_Ptr_List.Has_Next(Pkgs) loop
         Pkg_Class_Ptr_List.Get_Next(Pkgs, Pkg);
         -- Dump the *Hierarchy Tree* for the Flows with Pkg as Root
         Tree_Dot_Dump(Pkg.all);
         -- Dump *all the Information* about the Flows with Pkg as Root
         Pkg_Dot_Dump_Rec(Pkg.all);
      end loop;
   end World_Dot_Dump;

   -- F L O W  T R E E

   --
   -- Output a Hierarchy Subtree
   --
   procedure Tree_Dot_Dump (Flow: in Flow_Object'Class) is
      Filename: String := Make_Filename("PKG_CFG_Tree");
      My_File: File_Type;
   begin

      if Filename /= "" then
         declare
         begin
            Create(File=>My_File, Mode=>Out_File, Name=>Filename);
            Set_Output(My_File);
         exception
            when others =>
               Put_Line(Standard_Error, "[output] Unable to create file: " & Filename & " Redirecting to stdout.");
               Set_Output(Standard_Output);
         end;
      end if;

      Put("digraph ""PKG/CFG Tree"" {"); New_Line;
      Put("   label=""PKG/CFG Tree"";"); New_Line;
      Put("   node [shape=box];"); New_Line;
      Tree_Dot_Dump_Rec(Flow);
      Put("}"); New_Line;

      if Filename /= "" then
         Close(My_File);
         Set_Output(Standard_Output);
      end if;
   end Tree_Dot_Dump;

   --
   -- Output a Hierarchy Subtree recursively
   --
   procedure Tree_Dot_Dump_Rec (Flow: in Flow_Object'Class) is
      Flows:     Flow_Class_Ptr_List.Object;
      Next_Flow: Flow_Class_Ptr             := null;
      -- Prefix: String := Mask(U2s(Get_Prefix(Flow)));
   begin
      Flows := Get_Succs(Flow);
      if (Flow in Pkgs.Pkg_Object'Class) then
         -- This is a Package so we must handle it accordingly
         Put("   " & """P" & Id.To_String(Get_Id(Flow)) & """" &
             " [label=" & """" & Mask(Get_Name(Flow)));
         if (Flow in Pkgs.Body_Object'Class) then
            Put("\n(body)");
         else
            Put("\n(spec)");
         end if;
         --  if Prefix /= "" then
--              Put("\n in " & Prefix);
--           end if;
         Put(Mask(Refs   ("--REFS--",     Get_Refs(Flow))));
         Put(Mask(FDecls ("--FORMALS--",  Get_FDecls(Flow))));
         Put(Mask(RDecls ("--VARS--",     Get_Decls(Flow))));
         Put(Mask(Types  ("--TYPES--",    Get_Types(Flow))));
         Put(Mask(Aliases("--RENAMINGS--",Get_Rens(Flow))));
         Put("""" & "];");
         New_Line;
         -- Before writing the Succs -> also dump the Initialization Param Trees
         Write_Param_Trees("""P" & Id.To_String(Get_Id(Flow)) & """", Get_Decls(Flow));
         Write_Formal_Param_Trees("""P" & Id.To_String(Get_Id(Flow)) & """", Get_FDecls(Flow));
         Write_Type_Param_Trees("""P" & Id.To_String(Get_Id(Flow)) & """", Get_Types(Flow));
         -- Dump the Parameter Trees for the Renamings
         Write_Alias_Param_Trees("""P" & Id.To_String(Get_Id(Flow)) & """", Get_Rens(Flow));
         -- Write THE PTree
         if Get_PTree(Flow) /= null then
            Write_Param_Tree("""P" & Id.To_String(Get_Id(Flow)) & """", "Extra P-Tree",
                             Node_Object(Get_PTree(Flow).all)'Access);
         end if;
         New_Line;
      elsif (Flow in CFGs.CFG_Object'Class) then
         -- This is a CFG so we must handle it accordingly
         Put("   " & """G" & Id.To_String(Get_Id(Flow)) & """" &
             " [shape=ellipse,label=" & """" & Mask(Get_Name(Flow)));
         if (Flow in CFGs.Body_Object'Class) then
            Put("\n(body)");
         else
            Put("\n(spec)");
         end if;
         --  if Prefix /= "" then
--              Put("\n in " & Prefix);
--           end if;
         if Get_LName(CFG_Object(Flow)) /= "" then
            Put("\nLink: " & Get_LName(CFG_Object(Flow)));
         end if;
         Put(Mask(FDecls ("--FORMALS--",  Get_FDecls(Flow))));
         Put(Mask(RDecls ("--VARS--",     Get_Decls(Flow))));
         Put(Mask(Types  ("--TYPES--",    Get_Types(Flow))));
         Put(Mask(RDecls ("--PARAMS--",   Get_Params(CFG_Object(Flow)))));
         Put(Mask(Aliases("--RENAMINGS--",Get_Rens(Flow))));
         Put("""" & "];");
         New_Line;
         -- Before writing the Succs -> also dump the Initialization Param Trees
         Write_Param_Trees("""G" & Id.To_String(Get_Id(Flow)) & """", Get_Decls(Flow));
         Write_Formal_Param_Trees("""G" & Id.To_String(Get_Id(Flow)) & """", Get_FDecls(Flow));
         -- For CFGs also dump the Initialization Trees for their Parameters
         Write_Param_Trees("""G" & Id.To_String(Get_Id(Flow)) & """",
                           Get_Params(CFG_Object(Flow)));
         Write_Type_Param_Trees("""G" & Id.To_String(Get_Id(Flow)) & """", Get_Types(Flow));
         -- Dump the Parameter Trees for the Renamings
         Write_Alias_Param_Trees("""G" & Id.To_String(Get_Id(Flow)) & """", Get_Rens(Flow));
         -- Write THE PTree
         if Get_PTree(Flow) /= null then
            Write_Param_Tree("""G" & Id.To_String(Get_Id(Flow)) & """", "Extra P-Tree",
                             Node_Object(Get_PTree(Flow).all)'Access);
         end if;
         New_Line;
      else
         Put_Line(Standard_Error, "[DOT/OUTPUT] [ERR] Wrong Data Structure Type in the Flow Tree!");
         return;
      end if;
      -- Walk the Successors-List
      Flow_Class_Ptr_List.Reset(Flows);
      while Flow_Class_Ptr_List.Has_Next(Flows) loop
         Flow_Class_Ptr_List.Get_Next(Flows, Next_Flow);
         -- Make the correct Links
         if (Flow in Pkgs.Pkg_Object'Class) then
            Put("   " & """P");
         else
            Put("   " & """G");
         end if;
         Put(Id.To_String(Get_Id(Flow)) & """");
         Put(" -> ");
         if (Next_Flow.all in Pkgs.Pkg_Object'Class) then
            Put("""P");
         else
            Put("""G");
         end if;
         Put(Id.To_String(Get_Id(Next_Flow.all)) & """");
         Put(";"); New_Line;
         -- Proceed recursively
         Tree_Dot_Dump_Rec(Next_Flow.all);
      end loop;
   end Tree_Dot_Dump_Rec;

   -- Special List Output Methods

   --
   -- Compose a String out of a References List (with, use, ..)
   --
   function Refs(Name: in String; Refs_List: in Ref_Class_Ptr_List.Object) return String is
      Refs:    Ref_Class_Ptr_List.Object := Refs_List;
      Ref:     Ref_Class_Ptr             := null;
      Output:  Unbounded_String          := S2u("");
   begin
      if (Ref_Class_Ptr_List.Count(Refs) /= 0) then
         Output := S2u(Ada.Characters.Latin_1.LF & Name);
         Ref_Class_Ptr_List.Reset(Refs);
         while (Ref_Class_Ptr_List.Has_Next(Refs)) loop
            Output := Output & S2u(Ada.Characters.Latin_1.LF & "");
            Ref_Class_Ptr_List.Get_Next(Refs, Ref);
            if (Ref.all in With_Ref) then
               Output := Output & "W: ";
            elsif (Ref.all in Use_Pkg_Ref) then
               Output := Output & "UP: ";
            elsif (Ref.all in Use_Type_Ref) then
               Output := Output & "UT: ";
            end if;
            Output := Output & Unbounded_String'(Merge_UStr_List(Ref.Name,"."));
         end loop;
      end if;
      return U2s(Output);
   end Refs;

   --
   -- Compose a String out of an Alias List (Remaings, Instantiations, ..)
   --
   function Aliases (Name: in String; Alias_List: in Alias_Class_Ptr_List.Object) return String is
      Rens:        Alias_Class_Ptr_List.Object := Alias_List;
      Ren:         Alias_Class_Ptr             := null;
      Output:      Unbounded_String            := S2u("");
      First_Group: Boolean                     := True;
      Inst_Alias:  Instance_Alias_Class_Ptr    := null;
   begin
      if (Alias_Class_Ptr_List.Count(Rens) /= 0) then
         if (Name /= "") then
            Output := S2u(Ada.Characters.Latin_1.LF & Name);
         end if;
         Alias_Class_Ptr_List.Reset(Rens);
         while (Alias_Class_Ptr_List.Has_Next(Rens)) loop
            if (Name /= "") then
               Output := Output & S2u(Ada.Characters.Latin_1.LF & "");
            else
               if (not First_Group) then
                  Output := Output & S2u(",");
               end if;
            end if;
            Alias_Class_Ptr_List.Get_Next(Rens, Ren);
            First_Group := False;
            -- Compose the Alias Image
            Output := Output & Unbounded_String'(Merge_UStr_List(Ren.Old_Name, ".")) & S2u(" -> ") & Ren.New_Name;
            -- Check if this is an Instatiation Alias
            if (Ren.all in Instance_Alias'Class) then
               -- This is an Instantiation so we must save the Parameters as well
               Inst_Alias := Instance_Alias(Ren.all)'Access;
               Output := Output & S2u(" [") & Aliases("", Inst_Alias.Assocs) & S2u("] ");
            end if;
         end loop;
      end if;
      return U2s(Output);
   end Aliases;

   --
   -- Compose a String out of a Declaration List (Variables, Parameters, ..)
   --
   function RDecls(Name: in String; Decl_List: in Declaration_Ptr_List.Object) return String is
      Decls:  Declaration_Ptr_List.Object := Decl_List;
      Decl:   Declaration_Ptr             := null;
      Output: Unbounded_String            := S2u("");
   begin
      if (Declaration_Ptr_List.Count(Decls) /= 0) then
         Output := S2u(Ada.Characters.Latin_1.LF & Name);
         Declaration_Ptr_List.Reset(Decls);
         while (Declaration_Ptr_List.Has_Next(Decls)) loop
            Output := Output & S2u(Ada.Characters.Latin_1.LF & "");
            Declaration_Ptr_List.Get_Next(Decls, Decl);
            -- Write Variables and Type
            Output := Output & Unbounded_String'(Merge_UStr_List(Decl.Vars,",")) & ": " & Type_Image(Decl.VType);
         end loop;
      end if;
      return U2s(Output);
   end RDecls;

   --
   -- Compose a String out of a Formal Declaration List
   --
   function FDecls(Name: in String; Decl_List: in Formal_Decl_Class_Ptr_List.Object) return String is
      Decls:  Formal_Decl_Class_Ptr_List.Object := Decl_List;
      Decl:   Formal_Decl_Class_Ptr             := null;
      Output: Unbounded_String                  := S2u("");
   begin
      if (Formal_Decl_Class_Ptr_List.Count(Decls) /= 0) then
         Output := S2u(Ada.Characters.Latin_1.LF & Name);
         Formal_Decl_Class_Ptr_List.Reset(Decls);
         while (Formal_Decl_Class_Ptr_List.Has_Next(Decls)) loop
            Output := Output & S2u(Ada.Characters.Latin_1.LF & "");
            Formal_Decl_Class_Ptr_List.Get_Next(Decls, Decl);
            -- Write Variables and Type
            Output := Output & Unbounded_String'(Merge_UStr_List(Decl.Vars,",")) & ": " & Type_Image(Decl.VType);
            -- Write Aliases
            Output := Output & S2u(" [") & Aliases("", Decl.Assocs) & S2u("] ");
         end loop;
      end if;
      return U2s(Output);
   end FDecls;

   --
   -- Compose a String out of a Type List
   --
   function Types(Name: in String; Types_List: in Type_Descr_Ptr_List.Object) return String is
      Decls:  Type_Descr_Ptr_List.Object := Types_List;
      Decl:   Type_Descr_Ptr             := null;
      Output: Unbounded_String           := S2u("");
   begin
      if (Type_Descr_Ptr_List.Count(Decls) /= 0) then
         Output := S2u(Ada.Characters.Latin_1.LF & Name);
         Type_Descr_Ptr_List.Reset(Decls);
         while (Type_Descr_Ptr_List.Has_Next(Decls)) loop
            Type_Descr_Ptr_List.Get_Next(Decls, Decl);
            Output := Output & Ada.Characters.Latin_1.LF & Type_Image(Decl.all);
         end loop;
      end if;
      return U2s(Output);
   end Types;

   -- Special Param Tree Output Methods

   -- Dump a Parameter Tree
   procedure Write_Param_Tree (Prev: in String; New_Name: in String; Root: in Node_Class_Ptr) is
   begin
      if not Options.Param_Tree_Output then
         return;
      end if;

      if New_Name /= "" then
         -- Create a Pseudo Node with the Information in New_Name
         Put("    ""N"); Put(Id.To_String(Get_Id(Flow_Object(Root.all))) & "T");
         Put(""" [label="""); Put(New_Name); Put("""");
         Put(",style=dashed,color=blue];");
         New_Line;
         -- Link the previous Node to the Pseudo one
         Put(Prev); Put(" -> ""N"); Put(Id.To_String(Get_Id(Flow_Object(Root.all))) & "T"); Put(""";");
         New_Line;
         -- Link the Pseudo Node to the Root of the Param Tree
         Put("""N"); Put(Id.To_String(Get_Id(Flow_Object(Root.all))) & "T"); Put(""" -> ""N");
         Put(Id.To_String(Get_Id(Flow_Object(Root.all)))); Put(""";");
         New_Line;
      else
         -- Link the previous Node to the Root of the PTree
         Put(Prev); Put(" -> ""N"); Put(Id.To_String(Get_Id(Flow_OBject(Root.all)))); Put(""";");
      end if;
      -- Dump the Tree recursively
      Write_Node(Root.all, Parameters);
   end Write_Param_Tree;

   --
   -- Dump the Parameter Trees for a List of Declarations
   --
   procedure Write_Param_Trees (Name: in String; Decl_List: in Declaration_Ptr_List.Object) is
      Decls: Declaration_Ptr_List.Object := Decl_List;
      Decl:  Declaration_Ptr             := null;
   begin
      if not Options.Param_Tree_Output then
         return;
      end if;

      if (Declaration_Ptr_List.Count(Decls) /= 0) then
         Declaration_Ptr_List.Reset(Decls);
         while (Declaration_Ptr_List.Has_Next(Decls)) loop
            Declaration_Ptr_List.Get_Next(Decls, Decl);
            if (Decl.Init /= null) then
               Write_Param_Tree(Name, Merge_UStr_List(Decl.Vars, ",") & "\nINIT",
                                Node_Object(Decl.Init.all)'Access);
            else
                null;
            end if;
            if (Decl.VType.Descr_Tree /= null) then
               Write_Param_Tree(Name, Merge_UStr_List(Decl.Vars, ",") & "\nDECL/TYPE",
                                Node_Object(Decl.VType.Descr_Tree.all)'Access);
            else
               null;
            end if;
         end loop;
      end if;
   end Write_Param_Trees;

   --
   -- Dump the Parameter Trees for a List of Formal Declarations
   --
   procedure Write_Formal_Param_Trees (Name: in String; Decl_List: in Formal_Decl_Class_Ptr_List.Object) is
      Decls: Formal_Decl_Class_Ptr_List.Object := Decl_List;
      Decl:  Formal_Decl_Class_Ptr             := null;
   begin
      if not Options.Param_Tree_Output then
         return;
      end if;

      if (Formal_Decl_Class_Ptr_List.Count(Decls) /= 0) then
         Formal_Decl_Class_Ptr_List.Reset(Decls);
         while (Formal_Decl_Class_Ptr_List.Has_Next(Decls)) loop
            Formal_Decl_Class_Ptr_List.Get_Next(Decls, Decl);
            if (Decl.Init /= null) then
               Write_Param_Tree(Name, Merge_UStr_List(Decl.Vars, ",") & "\nINIT",
                                Node_Object(Decl.Init.all)'Access);
            end if;
            if (Decl.VType.Descr_Tree /= null) then
               Write_Param_Tree(Name, Merge_UStr_List(Decl.Vars, ",") & "\nFDECL/TYPE",
                                Node_Object(Decl.VType.Descr_Tree.all)'Access);
            end if;
         end loop;
      end if;
   end Write_Formal_Param_Trees;

   --
   -- Dump the Parameter Trees for a List of Aliases
   --
   procedure Write_Alias_Param_Trees (Name: in String; Alias_List: in Alias_Class_Ptr_List.Object) is
      Rens:        Alias_Class_Ptr_List.Object := Alias_List;
      Ren:         Alias_Class_Ptr             := null;
      Output:      Unbounded_String            := S2u("");
   begin
      if not Options.Param_Tree_Output then
         return;
      end if;

      if (Alias_Class_Ptr_List.Count(Rens) /= 0) then
         Alias_Class_Ptr_List.Reset(Rens);
         while (Alias_Class_Ptr_List.Has_Next(Rens)) loop
            Alias_Class_Ptr_List.Get_Next(Rens, Ren);
            if Ren.PTree /= null then
               Write_Param_Tree(Name, U2s(Ren.New_Name) & "\n(RENAME)",
                                Node_Object(Ren.PTree.all)'Access);
            end if;
         end loop;
      end if;
   end Write_Alias_Param_Trees;

   --
   -- Dump the Parameter Trees for a List of Type Descriptions
   --
   procedure Write_Type_Param_Trees (Name: in String; Types_List: in Type_Descr_Ptr_List.Object) is
      Decls: Type_Descr_Ptr_List.Object := Types_List;
      Decl:  Type_Descr_Ptr             := null;
   begin
      if not Options.Param_Tree_Output then
         return;
      end if;

      Type_Descr_Ptr_List.Reset(Decls);
      while (Type_Descr_Ptr_List.Has_Next(Decls)) loop
         Type_Descr_Ptr_List.Get_Next(Decls, Decl);
         if (Decl.Descr_Tree /= null) then
            Write_Param_Tree(Name, U2s(Decl.Type_Name) & "\nDESCR/TREE",
                             Node_Object(Decl.Descr_Tree.all)'Access);
         end if;
         Write_Param_Trees(Name, Decl.Components);
         Write_Param_Trees(Name, Decl.Discriminants);
      end loop;
   end Write_Type_Param_Trees;

   -- Special Types Methods

   --
   -- Construct a String Image of a Type Description Structure
   --
   function Type_Image (Type_Info: in Type_Descr) return String is
      Decl:   Type_Descr       := Type_Info;
      Output: Unbounded_String := S2u("");
   begin
      -- Write the Name of the Type
      Output := Output & Decl.Type_Name;
      -- Is it private?
      if (Decl.Is_Private) then
         Output := Output & S2u(" (private)");
      end if;
      -- Write the Supertype
      if (not UStr_List.empty(Decl.Super_Type)) then
         Output := Output & S2u(" {ST: " & Merge_UStr_List(Decl.Super_Type,".") & "}");
      end if;
      -- Write the Component Type
      if (not UStr_List.empty(Decl.Component_Type)) then
         Output := Output & S2u(" {CT: " & Merge_UStr_List(Decl.Component_Type,".") & "}");
      end if;
      -- Write the Found Idents
      if (not UStr_List_List.empty(Decl.Used_Idents)) then
         Output := Output & S2u(" {IDs: " & Merge_UStr_List_List(Decl.Used_Idents,".",",") & "}");
      end if;
      -- Write the Type Components
      Output := Output & RDecls("[COMPONENTS]", Decl.Components);
      Output := Output & RDecls("[DISCRIMINANTS]", Decl.Discriminants);
      -- Done
      return U2s(Output);
   end Type_Image;

   -- F L O W  S T R U C T U R E S

   -- N O D E S

   --
   -- Write a Node and follow its Succs recursively
   --
   procedure Write_Node (Cur: in out Node_Object'Class; Cfg_Class: in Cfg_Type) is
      Succs:               Flow_Class_Ptr_List.Object;
      Succ:                Node_Class_Ptr             := null;
      Flow,Dest_Ptr:                Flow_Class_Ptr             := null;
      Dest_Name, Var_Name,
        Type_Name:         Unbounded_String           := S2u("");
   begin
      -- output node type
      Put("    ""N");
      Put(Id.To_String(Get_Id(Cur)));
      Put(""" [label=""");
      if (Cur in Call_Node_Object) or
        (Cur in Abort_Node_Object) or
        (Cur in Entry_Call_Node_Object) or
        (Cur in Param_Call_Node_Object) or
        (Cur in Accept_Node_Object) then
         if (Cur in Abort_Node_Object) then
            Put("Abort Node");
         elsif (Cur in Entry_Call_Node_Object) then
            Put("Entry Call Node");
         elsif (Cur in Accept_Node_Object) then
            Put("Accept Node");
         elsif (Cur in Param_Call_Node_Object) then
            Put("Parameter Call Node ");
         else
            Put("Call Node ");
         end if;
         Put(Id.To_String(Get_Id(Cur)) & ":\n");
         Dest_Name := S2u(Get_Dest_Name(Call_Node_Object(Cur)));
         if U2s(Dest_Name) /= "" then
            Put("Dest: " & U2s(Mask(Dest_Name)) & "\n");
         end if;

         Dest_Ptr := Get_Call_Destination(Call_Node_Object(Cur));
         if Dest_Ptr /= null then
            Put("Dest_Ptr_Name: " & Get_Name(Flow_Object(Dest_Ptr.all)) & "\n");
         else
            Put("Dest_Ptr_Name: null \n");
         end if;

      elsif Cur in Goto_Jump_Node_Object then
         Put("Goto Jump Node ");
         Put(Id.To_String(Get_Id(Cur)) & ":\n");
         if Get_Target(Goto_Jump_Node_Object(Cur)) /= "" then
            Put("Target: " & Get_Target(Goto_Jump_Node_Object(Cur)));
         end if;
      elsif Cur in Trivial_Exit_Jump_Node_Object then
         Put("T-Exit Jump Node ");
         Put(Id.To_String(Get_Id(Cur)) & ":\n");
         if Get_Target(Trivial_Exit_Jump_Node_Object(Cur)) /= "" then
            Put("Target: " & Get_Target(Trivial_Exit_Jump_Node_Object(Cur)));
         end if;
      elsif Cur in Complex_Exit_Jump_Node_Object then
         Put("C-Exit Jump Node ");
         Put(Id.To_String(Get_Id(Cur)) & ":\n");
         if Get_Target(Complex_Exit_Jump_Node_Object(Cur)) /= "" then
            Put("Target: " & Get_Target(Complex_Exit_Jump_Node_Object(Cur)));
         end if;
      elsif Cur in Param_Node_Object then
         Put("Parameter Node ");
         Put(Id.To_String(Get_Id(Cur)) & ":\n");
         Var_Name := S2u(Get_Var_Name(Param_Node_Object(Cur)));
         if (Var_Name /= "") then
            Put("Variable: " & U2s(Var_Name));
         end if;
      elsif Cur in Param_Alloc_Node_Object then
         Put("Parameter Allocation Node ");
         Put(Id.To_String(Get_Id(Cur)) & ":\n");
         Type_Name := S2u(Get_Type_Name(Param_Alloc_Node_Object(Cur)));
         if (Type_Name /= "") then
            Put("Type: " & U2s(Type_Name));
         end if;
      elsif Cur in Assign_Node_Object then
         Put("Assignment Node ");
         Put(Id.To_String(Get_Id(Cur)) & ":\n");
      elsif Cur in Return_Node_Object then
         Put("Return Node ");
         Put(Id.To_String(Get_Id(Cur)) & ":\n");
      elsif Cur in Terminate_Node_Object then
         Put("Terminate Node ");
         Put(Id.To_String(Get_Id(Cur)) & ":\n");
      elsif Cur in Finite_Loop_Node_Object then
         Put("Finite Loop Node ");
         Put(Id.To_String(Get_Id(Cur)) & ":\n");
      elsif Cur in Infinite_Loop_Node_Object then
         Put("Infinite Loop Node ");
         Put(Id.To_String(Get_Id(Cur)) & ":\n");
      elsif Cur in Branch_Node_Object then
         Put("Branch Node ");
         Put(Id.To_String(Get_Id(Cur)) & ":\n");
      else
         Put("Node ");
         Put(Id.To_String(Get_Id(Cur)) & ":\n");
      end if;
      -- Write the Node Name
      if (String'(Get_Name(Cur)) /= "") then
         Put("\nName: " & Get_Name(Cur));
      end if;
      -- Write the Node Identifiers
      if (not UStr_List_List.Empty(Get_Idents(Cur))) then
         Put("\nIdents: " & Mask(Merge_UStr_List_List(Get_Idents(Cur),".",",")));
      end if;
      -- Dump eventual Declarations
      Put(Mask(RDecls("--DECLS--",Get_Decls(Flow_Object(Cur)))));
      -- In dot Format "s and \s need to be masked with \
      Put("\n" & Mask(Get_Info(Cur)));
      Put("""");
      -- Make the Node dashed if this is a Parameter Tree
      if (Cfg_Class = Parameters) then
         Put(",style=dashed");
      end if;
      -- Change the Color of the Node
      if (Cur in Call_Node_Object) or
        (Cur in Param_Call_Node_Object) then
         Put(",color=red");
      elsif Cur in Jump_Node_Object then
         Put(",color=blue");
      elsif (Cur in Accept_Node_Object) or
        (Cur in Entry_Call_Node_Object) or
        (Cur in Abort_Node_Object)
      then
         Put(",color=green");
      end if;
      -- Done
      Put("];");
      New_Line;
      -- Mark the Node as visited
      Set_Mark(Cur, 1);

      -- Dump the LHS-Tree for an Assignment Node
      if (Cur in Assign_Node_Object) then
         if (Get_LHS_Tree(Assign_Node_Object(Cur)) /= null) then
            Succ := Get_LHS_Tree(Assign_Node_Object(Cur));
            Write_Param_Tree("""N" & Id.To_String(Get_Id(Cur)) & """", "", Succ);
         end if;
      end if;

      -- Dump the RHS-Tree (if any)
      if (Get_RHS_Tree(Cur) /= null) then
         Succ := Get_RHS_Tree(Cur);
         Write_Param_Tree("""N" & Id.To_String(Get_Id(Cur)) & """", "", Succ);
      end if;

      -- Output Edges and start/continue Recursion
      Succs := Get_Succs(Cur);

      -- Follow Successors
      Flow_Class_Ptr_List.Reset(Succs);
      while Flow_Class_Ptr_List.Has_Next(Succs) loop
         Flow_Class_Ptr_List.Get_Next(Succs, Flow);
         Succ := Node_Class_Ptr(Flow);
         Put("    ""N");
         Put(Id.To_String(Get_Id(Cur)));
         Put(""" -> ""N");
         Put(Id.To_String(Get_Id(Succ.all)));
         Put(""";");
         New_Line;
         if Get_Mark(Succ.all) /= 1 then
            Write_Node(Succ.all, Cfg_Class);
         end if;
      end loop;

   end Write_Node;

   -- C F G S

   --
   -- Output a dot Representation of a CFG
   --
   procedure Cfg_Dot_Dump (Cfg: in CFG_Object'Class; Cfg_Class: in Cfg_Type) is
      Name: String         := Compose_Hierarchical_Name(Cfg);
      Root: Node_Class_Ptr := Get_Head(Cfg);
      Filename: String := Make_Filename(Name);
      My_File: File_Type;
   begin

      -- (check if) we need to open a file
      if Filename /= "" then
         declare
         begin
            Create(File=>My_File, Mode=>Out_File, Name=>Filename);
            Set_Output(My_File);
         exception
            when others =>
               Put_Line(Standard_Error, "[output] Unable to create file: " & Filename & " Redirecting to stdout.");
               Set_Output(Standard_Output);
         end;
      end if;

      -- begin digraph
      Put_Line("digraph """ & Mask(Name) & """ {");
      Put_Line("    label=""" & Mask(Name) & """;");
      Put_Line("    node [shape=box];");
      New_Line;
      -- add nodes/edges
      if (Root /= null) then
         Write_Node(Root.all, Cfg_Class);
      end if;
      -- end digraph
      Put_Line("}");
      New_Line;

      if Filename /= "" then
         Close(My_File);
         -- reset output to stdout
         Set_Output(Standard_Output);
      end if;

   end Cfg_Dot_Dump;

   -- P A C K A G E S

   --
   -- Output the CFGs in all the Packages in this Tree (recursively)
   --
   procedure Pkg_Dot_Dump_Rec (Flow: in Flow_Object'Class) is
      Flows:     Flow_Class_Ptr_List.Object;
      Next_Flow: Flow_Class_Ptr             := null;
   begin
      if (Flow in CFG_Object'Class) then
         Cfg_Dot_Dump(CFG_Object(Flow), Normal);
      elsif (Flow in Pkg_Object'Class) then
         null;
      else
         Put_Line(Standard_Error, "[DOT/OUTPUT] [ERR] Wrong Data Structure Type in the Flow Tree!");
         return;
      end if;
      Flows := Get_Succs(Flow);
      Flow_Class_Ptr_List.Reset(Flows);
      while Flow_Class_Ptr_List.Has_Next(Flows) loop
         Flow_Class_Ptr_List.Get_Next(Flows, Next_Flow);
         Pkg_Dot_Dump_Rec(Next_Flow.all);
      end loop;
   end Pkg_Dot_Dump_Rec;

end Cfg_To_Dot.Output;
